class Block {
  //the parent class for all blocks
  
  //global variables
  float x, y;
  PImage img;
  int x_list_pos, y_list_pos;
  String type;
  Item item;
  float block_health, max_block_health;
  int mining_level;
  PImage[] block_breaking_imgs = new PImage[4];
  Light light;
  float blockBrightness;
  color currentLightColor;
  boolean collisionPossible;
  float level = -1; //lava and water blocks
  
  Block(float _x, float _y, int _x_list_pos, int _y_list_pos, String img_name, String _type, int _mining_level, Light _light) {
    //init global variables
    x = _x;
    y = _y;
    img = texture_manager.get_texture(img_name);
    x_list_pos = _x_list_pos;
    y_list_pos = _y_list_pos;
    type = _type;
    max_block_health = 100;
    block_health = max_block_health;
    mining_level = _mining_level;
    //the block-breaking images are a mostly transparent PNG with cracks of various length - it is
    //utilized to make a block-breaking animation without having to make 4 different images of the same
    //block in various states of deterioration.
    for (int i = 0; i < block_breaking_imgs.length; i++) {
      block_breaking_imgs[i] = texture_manager.get_texture("block_breaking_" + str(i+1) + ".png");
    }
    light = _light;
    collisionPossible = true;
  }
  
  void update(ArrayList<Light> lights) {
    if (light == null) {
      roughLighting(lights);
    }
    //display(lights);
    if (!type.equals("lava") && level != -1) {
      exit();
    }
  }
  
  void roughLighting(ArrayList<Light> lights) {
    //rough lightning function
    blockBrightness = 0;
    float temp_r = 0;
    float temp_g = 0;
    float temp_b = 0;
    int light_amount = 0;
    for (Light light : lights) {
      float dist_to_light = dist(x, y, light.x, light.y);
      float current_light_intensity = constrain(light.intensity/pow(dist_to_light/50, 2), 0, light.intensity);
      float tempBrightness = current_light_intensity/10;
      blockBrightness = max(blockBrightness, tempBrightness);
      if (dist_to_light < light.intensity) {
        temp_r += red(light.lightColor)*current_light_intensity;
        temp_g += green(light.lightColor)*current_light_intensity;
        temp_b += blue(light.lightColor)*current_light_intensity;
        light_amount += current_light_intensity;
      }
    }
    temp_r /= light_amount;
    temp_g /= light_amount;
    temp_b /= light_amount;
    currentLightColor = color(temp_r, temp_g, temp_b);
  }
  
  void display(ArrayList<Light> lights) {
    //displays the image and checks if blitting a block-breaking animation is necessary
    imageMode(CENTER);
    if (!type.equals("lava") && light == null) {
      tint(currentLightColor);
    }
    image(img, x, y);
    noTint();
    if (block_health > 0 && block_health <= max_block_health*0.2) {
      image(block_breaking_imgs[3], x, y);
    } else if (block_health > max_block_health*0.2 && block_health <= max_block_health*0.4) {
      image(block_breaking_imgs[2], x, y);
    } else if (block_health > max_block_health*0.4 && block_health <= max_block_health*0.6) {
      image(block_breaking_imgs[1], x, y);
    } else if (block_health > max_block_health*0.6 && block_health <= max_block_health*0.8) {
      image(block_breaking_imgs[0], x, y);
    }
    if (!type.equals("lava") && light == null) {
      rectMode(CENTER);
      fill(0, 0, 0, 255-blockBrightness*25);
      rect(x, y, 20, 20);
    }
  }
  
  void kill() {
    blocks[x_list_pos][y_list_pos] = null;
  }
}