class Enemy {
  //global variables
  PVector location, velocity, knockback;
  float speed;
  PImage img;
  String type;
  Block[][] blocks;
  
  float g_const;
  boolean knockback_on, touching_ground;
  boolean facing_right;
  boolean movingRight, movingLeft;
  boolean is_dead;
  int attack, health, max_health, attack_cooldown, max_attack_cooldown;
  float jump;
  
  int dis = 20;
  int enemy_offset = 15;
  
  Enemy(float _x, float _y, String _type, String img_name, int _attack_cooldown, Block[][] _blocks) {
    //global variables init
    location = new PVector(_x, _y);
    velocity = new PVector(0, 0);
    knockback = new PVector(0, 0);
    knockback_on = true;
    img = texture_manager.get_texture(img_name);
    facing_right = true;
    blocks = _blocks;
    type = _type;
    
    g_const = 1;
    touching_ground = false;
    
    //attack_cooldown = 0;
    attack_cooldown = _attack_cooldown;
    max_attack_cooldown = _attack_cooldown;
    is_dead = false;
    jump = 15;
  }
  
  void update(Player _player) {
    health_check();
    checkAttack(_player);
    move();
    if (!knockback_on) {
      applyGravity();
    }
    display();
  }
  
  void move() {
    if (player.x < location.x) {
      movingLeft = true;
      movingRight = false;
    } else {
      movingLeft = false;
      movingRight = true;
    }
    
    if (knockback_on) {
      boolean noCollision = true;
      for (int i = max(0, round(location.x/20)-dis+enemy_offset); i < min(round(location.x/20)+dis+enemy_offset, blocks.length); i++) {
        for (int j = max(0, round(location.y/20)-dis+enemy_offset); j < min(round(location.y/20)+dis+enemy_offset, blocks[i].length); j++) {
          if (blocks[i][j] != null && blocks[i][j].collisionPossible && detectBlockCollision(blocks[i][j])) {
            noCollision = false;
          }
        }
      }
      if (!noCollision) {
        knockback_on = false;
        knockback = new PVector(0, 0);
        boolean collision = true;
        while (collision) {
          collision = false;
          for (int i = max(0, round(location.x/20)-dis+enemy_offset); i < min(round(location.x/20)+dis+enemy_offset, blocks.length); i++) {
            for (int j = max(0, round(location.y/20)-dis+enemy_offset); j < min(round(location.y/20)+dis+enemy_offset, blocks[i].length); j++) {
              if (blocks[i][j] != null && blocks[i][j].collisionPossible && detectBlockCollision(blocks[i][j])) {
                collision = true;
              }
            }
          }
          location.y--;
        }
      } else {
        knockback.y += g_const;
        location.add(knockback);
      }
    } else if (movingLeft) {
      location.x -= speed;
      boolean noCollision = true;
      for (int i = max(0, round(location.x/20)-dis+enemy_offset); i < min(round(location.x/20)+dis+enemy_offset, blocks.length); i++) {
        for (int j = max(0, round(location.y/20)-dis+enemy_offset); j < min(round(location.y/20)+dis+enemy_offset, blocks[i].length); j++) {
          if (blocks[i][j] != null && blocks[i][j].collisionPossible && detectCollision(blocks[i][j])) {
            noCollision = false;
          }
        }
      }
      if (!noCollision) {
        location.x += speed;
        if (touching_ground) {
          velocity.y = -7;
        }
      }
    } else if (movingRight) {
      location.x += speed;
      boolean noCollision = true;
      for (int i = max(0, round(location.x/20)-dis+enemy_offset); i < min(round(location.x/20)+dis+enemy_offset, blocks.length); i++) {
        for (int j = max(0, round(location.y/20)-dis+enemy_offset); j < min(round(location.y/20)+dis+enemy_offset, blocks[i].length); j++) {
          if (blocks[i][j] != null && blocks[i][j].collisionPossible && detectCollision(blocks[i][j])) {
            noCollision = false;
          }
        }
      }
      if (!noCollision) {
        location.x -= speed;
        if (touching_ground) {
          velocity.y = -7;
        }
      }
    }
    if ((player.y + img.height*3 < location.y) && touching_ground) {
      velocity.y = -jump;
    }
  }
  
  void display() {
    image(img, location.x, location.y);
  }
  
  void applyGravity() {
    velocity.y += g_const;
    location.y += velocity.y;
    boolean noCollision = true;
    touching_ground = false;
    for (int i = max(0, round(location.x/20)-dis+enemy_offset); i < min(round(location.x/20)+dis+enemy_offset, blocks.length); i++) {
      for (int j = max(0, round(location.y/20)-dis+enemy_offset); j < min(round(location.y/20)+dis+enemy_offset, blocks[i].length); j++) {
        if (blocks[i][j] != null && blocks[i][j].collisionPossible && detectBlockCollision(blocks[i][j])) {
          noCollision = false;
        }
      }
    }
    if (!noCollision) {
      location.y -= velocity.y;
      velocity.y = 0;
      touching_ground = true;
    }
  }
  
  boolean detectBlockCollision (Block block) {
    if  ((location.x + img.width/2 >= block.x - block.img.width/2) &&
         (location.x - img.width/2 <= block.x + block.img.width/2) &&
         (location.y + img.height/2 >= block.y - block.img.height/2) &&
         (location.y - img.height/2 <= block.y + block.img.height/2))
    {
          return true;
    }
    return false;
  }
  
  boolean detectCollision (Block block) {
    if  ((location.x + img.width/2 >= block.x - block.img.width/2) &&
         (location.x - img.width/2 <= block.x + block.img.width/2) &&
         (location.y + img.height/2 >= block.y - block.img.height/2) &&
         (location.y - img.height/2 <= block.y + block.img.height/2))
    {
          return true;
    }
    return false;
  }
  
  void damage(float player_attack) {
    health -= player_attack;
  }
  
  void health_check() {
    if (health <= 0) {
      is_dead = true;
    } else if (location.y > 10000) {
      is_dead = true;
    }
  }
  
  void checkAttack(Player _player) {
    if (attack_cooldown == 0) {
      attack_cooldown = max_attack_cooldown;
      if (location.x > _player.x - _player.img_right.width/2 - img.width/2 && location.x < _player.x + _player.img_right.width/2 + img.width/2 &&
          location.y > _player.y - _player.img_right.height/2 - img.height/2 && location.y < _player.y + _player.img_right.height/2 + img.height/2) {
            _player.damage(attack);
      }
    } else {
      attack_cooldown--;
    }
  }
}