class Game {
  //the class for loading an updating the world
  
  //global variables
  int load_counter;
  float block_yoff, block_x, roughness, size;
  float block_update_counter, max_block_update_counter;
  String state;
  PImage[] backgrounds, biome_backgrounds;
  PImage main_menu_screen;
  PImage loading_bar, original_loading_bar;
  ArrayList<Particle> particles;
  ArrayList<Enemy> enemies;
  ArrayList<EnemyBullet> enemy_bullets;
  FloatList recent_fps;
  boolean underground_init, desert_init, wasteland_init;
  int mountain_caves_amount;
  float[] mountain_caves_height;
  float spawn_speed;
  int player_offset;
  int forest_enemies, cave_enemies, desert_enemies, wasteland_enemies;
  
  Game() {
    //prints instructions
    println("Welcome to Epic of Xathoebar! This is a very rough draft of the Major Project Assignment,");
    println("this version mainly dealing with world generation.");
    println("\nControls:");
    println("WAD to move\nE to access inventory\nLeft-Click to break blocks\nRight-Click to place blocks.");
    println("\nThe game might take a while to load, so please enjoy the beautiful loading bar in the meantime!");
    println("\nRemember to use recipies.txt and tools.txt to access the recipies and mining levels for the items in the game!");
    
    //setup
    noStroke();
    textFont(createFont("Arial", 15, true), 15);
    textAlign(RIGHT);
    rectMode(CENTER);
    imageMode(CENTER);
    
    //global variables init
    backgrounds = new PImage[4];
    for (int i = 0; i < backgrounds.length; i++) {
      backgrounds[i] = texture_manager.get_texture("background_"+(i+1)+".png");
    }
    biome_backgrounds = new PImage[4];
    String[] biome_background_names = {"forest", "cave", "desert", "wasteland"};
    for (int i = 0; i < biome_backgrounds.length; i++) {
      biome_backgrounds[i] = texture_manager.get_texture(biome_background_names[i] + "_background.png");
    }
    main_menu_screen = texture_manager.get_texture("main_menu.png");
    roughness = 10;
    size = 2;
    original_loading_bar = texture_manager.get_texture("loading_bar.png");
    loading_bar = original_loading_bar.get();
    block_x = -land_width/2;
    block_yoff = 0;
    //state = "load_world_file";
    state = "menu";
    load_counter = 0;
    max_block_update_counter = 45;
    block_update_counter = max_block_update_counter;
    particles = new ArrayList<Particle>();
    enemies = new ArrayList<Enemy>();
    enemy_bullets = new ArrayList<EnemyBullet>();
    recent_fps = new FloatList();
    underground_init = false;
    desert_init = false;
    wasteland_init = false;
    mountain_caves_amount = round(random(3, 5));
    mountain_caves_height = new float[mountain_caves_amount];
    for (int i = 0; i < mountain_caves_amount; i++) {
      mountain_caves_height[i] = random(1, 3);
    }
    spawn_speed = 0.2;
    player_offset = 20;
    forest_enemies = 0;
    cave_enemies = 0;
    desert_enemies = 0;
    wasteland_enemies = 0;
  }

  void loadGame(int loadingState) {
    //generate the world terrain
    if (loadingState == 1) {    
      state = "load_2";
    } else if (loadingState == 2) {
      background(255);
      if (load_counter < land_width) {
        int i = load_counter;
        if (load_counter < biome_width) { //load_counter < biome_width
          biome_type = "forest";
        } else if (load_counter < biome_width*2) {
          biome_type = "underground";
        } else if (load_counter < biome_width*3) {
          biome_type = "desert";
        } else {
          biome_type = "wasteland";
        }
        if (biome_type.equals("forest")) {
          block_yoff += 0.05;
          //uses perlin noise to make the shape
          //float n = round(noise(block_yoff) * height / 20) * 20 + 2000; //2000 - 2720
          float n = round(map(noise(block_yoff), 0, 1, 2000, 2720)/20)*20;
          //float n1 = ((round(noise(block_yoff*2) * height / 20)+10) * 20) + 2000;
          float n1 = round(map(noise(block_yoff*2), 0, 1, 2200, 2920)/20)*20;
          //float n2 = ((round(noise(block_yoff*1.5) * height / 20)+30) * 20) + 2000;
          float n2 = round(map(noise(block_yoff*1.5), 0, 1, 2600, 3320)/20)*20;
          //float n3 = ((round(noise(block_yoff*2) * height / 20)+50) * 20) + 2000;
          float n3 = round(map(noise(block_yoff*2), 0, 1, 3000, 3720)/20)*20;
          //float n4 = ((round(noise(block_yoff*3) * height / 20)+70) * 20) + 2000;
          float n4 = round(map(noise(block_yoff*3), 0, 1, 3400, 4120)/20)*20;
          for (float j = n; j < max(n+20, n1); j += 20) {
            blocks[round(i)][round(j/20)] = new DirtBlock(block_x, j, round(i), round(j/20));
            background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "dirt_bw.png", "dirt bw");
          }
          for (float j = max(n+20, n1); j < max(max(n+20, n1)+20, n2); j += 20) {
            blocks[round(i)][round(j/20)] = new CustomBlock(round(i), round(j/20), "stoney block");
            background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "stoney_bw.png", "stoney bw");
          }
          for (float j = max(max(n+20, n1)+20, n2); j < max(max(max(n+20, n1)+20, n2)+20, n3); j += 20) {
            blocks[round(i)][round(j/20)] = new CustomBlock(round(i), round(j/20), "stone block");
            background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "stone_bw.png", "stone bw");
          }
          for (float j = max(max(max(n+20, n1)+20, n2)+20, n3); j < max(max(max(max(n+20, n1)+20, n2)+20, n3)+20, n4); j += 20) {
            blocks[round(i)][round(j/20)] = new CustomBlock(round(i), round(j/20), "granite block");
            background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "granite_bw.png", "granite bw");
          }
          for (float j = max(max(max(max(n+20, n1)+20, n2)+20, n3)+20, n4); j < max(max(max(max(n+20, n1)+20, n2)+20, n3)+20, n4)+600; j += 20) {
            blocks[round(i)][round(j/20)] = new CustomBlock(round(i), round(j/20), "basalt block");
            background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "basalt_bw.png", "basalt bw");
          }
          saved_alt = n;
        } else if (biome_type.equals("underground")) {
          block_yoff += 0.007;
          //float n = round((noise(block_yoff)-0.5)*8 * height / 20) * 20 + 2000;
          float n = round(map(noise(block_yoff), 0, 1, -880, 4880)/20)*20;
          //float n1 = ((round(noise(block_yoff*2) * height / 20)+50) * 20) + 2000;
          //float n2 = ((round(noise(block_yoff*3) * height / 20)+70) * 20) + 2000;
          float n1 = round(map(noise(block_yoff*2), 0, 1, 3000, 3720)/20)*20;
          float n2 = round(map(noise(block_yoff*3), 0, 1, 3400, 4120)/20)*20;
          while (!underground_init) {
            block_yoff += 0.001;
            n = round((noise(block_yoff)-0.5)*8 * height / 20) * 20 + 2000;
            if (n == saved_alt) {
              underground_init = true;
            }
          }
          float[] ns = new float[mountain_caves_amount];
          for (int j = 0; j < mountain_caves_amount; j++) {
            ns[j] = (round((noise(block_yoff*mountain_caves_height[j])-0.5)*8 * height / 20)+((j+1)/(ns.length+1))*100) * 20 + 2000;
          }
          
          //stone layer
          for (float j = n; j < max(n+20, n1); j += 20) {
            try {
              blocks[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new CustomBlock(constrain(round(i), 0, blocks.length), constrain(round(j/20), 0, blocks[0].length), "stone block");
              //blocks[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new SandBlock(block_x, j, constrain(round(i), 0, blocks.length), constrain(round(j/20), 0, blocks[0].length));
              background_walls[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "stone_bw.png", "stone bw");  
            } catch (ArrayIndexOutOfBoundsException e) {
              
            }
          }
          
          //granite layer
          for (float j = max(n+20, n1); j < max(max(n+20, n1)+20, n2); j += 20) {
            try {
              blocks[round(i)][round(j/20)] = new CustomBlock(round(i), round(j/20), "granite block");
              //blocks[round(i)][round(j/20)] = new SandBlock(block_x, j, round(i), round(j/20));
              background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "granite_bw.png", "granite bw");
            } catch (ArrayIndexOutOfBoundsException e) {
              
            }
          }
          
          //basalt layer
          for (float j = max(max(n+20, n1)+20, n2); j < max(max(n+20, n1)+20, n2)+600; j += 20) {
            try {
              blocks[round(i)][round(j/20)] = new CustomBlock(round(i), round(j/20), "basalt block");
              //blocks[round(i)][round(j/20)] = new SandBlock(block_x, j, round(i), round(j/20));
              background_walls[round(i)][round(j/20)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "basalt_bw.png", "basalt bw");
            } catch (ArrayIndexOutOfBoundsException e) {
              
            }
          }
          
          //caves
          for (float ns_entry : ns) {
            for (float j = ns_entry; j < ns_entry+200; j+= 20) {
              try {
                blocks[round(i)][round(j/20)] = null;
                if (j < n) {
                  background_walls[round(i)][round(j/20)] = null;
                }
              } catch (ArrayIndexOutOfBoundsException e) {
              
              }
            }
          }
          saved_alt = n;
        }
        else if (biome_type.equals("desert")) {
          block_yoff += 0.003;
          float n = round(map(noise(block_yoff), 0, 1, -880, 4880)/20)*20;
          //float n1 = round(map(noise(block_yoff), 0, 1, -680, 4880)/20)*20;
          while (!desert_init) {
            block_yoff += 0.001;
            n = round((noise(block_yoff)-0.5)*8 * height / 20) * 20 + 2000;
            if (n == saved_alt) {
              desert_init = true;
            }
          }
          
          //sand layer
          for (float j = n; j < n+2000; j += 20) {
            try {
              blocks[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new SandBlock(block_x, j, constrain(round(i), 0, blocks.length), constrain(round(j/20), 0, blocks[0].length));
              //blocks[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new SandBlock(block_x, j, constrain(round(i), 0, blocks.length), constrain(round(j/20), 0, blocks[0].length));
              background_walls[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "sand_bw.png", "sand bw");  
            } catch (ArrayIndexOutOfBoundsException e) {
              
            }
          }
          saved_alt = n;
        }
        else if (biome_type.equals("wasteland")) {
          block_yoff += 0.01;
          float n = round(map(noise(block_yoff), 0, 1, -880, 4880)/20)*20;
          //float n1 = round(map(noise(block_yoff), 0, 1, -680, 4880)/20)*20;
          while (!wasteland_init) {
            block_yoff += 0.001;
            n = round((noise(block_yoff)-0.5)*8 * height / 20) * 20 + 2000;
            if (n == saved_alt) {
              wasteland_init = true;
            }
          }
          
          //dried lava layer
          for (float j = n; j < n+2000; j += 20) {
            try {
              blocks[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new DriedLavaBlock(block_x, j, constrain(round(i), 0, blocks.length), constrain(round(j/20), 0, blocks[0].length));
              //blocks[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new SandBlock(block_x, j, constrain(round(i), 0, blocks.length), constrain(round(j/20), 0, blocks[0].length));
              background_walls[constrain(round(i), 0, blocks.length)][constrain(round(j/20), 0, blocks[0].length)] = new CustomBackgroundWall(block_x, j, round(i), round(j/20), "dried_lava_bw.png", "dried lava bw");  
            } catch (ArrayIndexOutOfBoundsException e) {
              
            }
          }
        }
        block_x += 20;
        load_counter++;
        imageMode(CENTER);

        //a fancy loading bar to make things more interesting
        float transparency_counter = 0;
        if (i*100/land_width < 20) {
          image(backgrounds[0], width/2, height/2);
          transparency_counter = 0;
        } else if (i*100/land_width >= 20 && i*100/land_width < 30) {
          transparency_counter = (i*100/land_width-20);
          image(backgrounds[1], width/2, height/2);
          tint(255, 255-25*transparency_counter);
          //tint(255, 120);
          image(backgrounds[0], width/2, height/2);
        }

        if (i*100/land_width < 45 && i*100/land_width >= 30) {
          image(backgrounds[1], width/2, height/2);
          transparency_counter = 0;
        } else if (i*100/land_width >= 45 && i*100/land_width < 55) {
          transparency_counter = (i*100/land_width-45);
          image(backgrounds[2], width/2, height/2);
          tint(255, 255-25*transparency_counter);
          //tint(255, 120);
          image(backgrounds[1], width/2, height/2);
        }
        if (i*100/land_width < 70 && i*100/land_width >= 55) {
          image(backgrounds[2], width/2, height/2);
          transparency_counter = 0;
        } else if (i*100/land_width >= 70 && i*100/land_width < 80) {
          transparency_counter = (i*100/land_width-70);
          image(backgrounds[3], width/2, height/2);
          tint(255, 255-25*transparency_counter);
          //tint(255, 120);
          image(backgrounds[2], width/2, height/2);
        }

        if (i*100/land_width >= 80) {
          image(backgrounds[3], width/2, height/2);
          transparency_counter = 0;
        }

        noTint();
        image(loading_bar, width/2, height/2);
        rectMode(CORNER);
        fill(0);
        rect(width/2+loading_bar.width/2, height/2-loading_bar.height/2, -loading_bar.width+round(i*loading_bar.width/land_width), loading_bar.height);
        fill(255, 0, 0);
        textFont(createFont("Arial", 50, true), 50);
        textAlign(CENTER);
        text(i*100/land_width+"%", width/2, height/2+17);
        
        if (load_counter == land_width) {
          state = "load_3";
        }
      }
    } else {
      //ore veins
      
      //copper ore
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("stoney block") && random(0, 1) <= 0.005) {
            blocks[i][j] = new CustomBlock(i, j, "copper ore");
            float x_offset = 0;
            float y_offset = 0;
            for (int k = 0; k < random(4, 12); k++) {
              //x_offset += round(random(0, 2)-0.5)*2-1;
              //y_offset += round(random(0, 2)-0.5)*2-1;
              x_offset += round(random(-1, 1));
              y_offset += round(random(-1, 1));
              if (i+round(x_offset) < land_width && i+round(x_offset) > 0 && j+round(y_offset) < land_height && j+round(y_offset) > 0 && blocks[i+round(x_offset)][j+round(y_offset)] != null && blocks[i+round(x_offset)][j+round(y_offset)].type.equals("stoney block")) {
                blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "copper ore");
              }
            }
          }
        }
        block_x += 20;
      }
      //tin ore
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("stoney block") && random(0, 1) <= 0.005) {
            blocks[i][j] = new CustomBlock(i, j, "tin ore");
            float x_offset = 0;
            float y_offset = 0;
            for (int k = 0; k < random(4, 12); k++) {
              //x_offset += round(random(0, 2)-0.5)*2-1;
              //y_offset += round(random(0, 2)-0.5)*2-1;
              x_offset += round(random(-1, 1));
              y_offset += round(random(-1, 1));
              if (i+round(x_offset) < land_width && i+round(x_offset) > 0 && j+round(y_offset) < land_height && j+round(y_offset) > 0 && blocks[i+round(x_offset)][j+round(y_offset)] != null && blocks[i+round(x_offset)][j+round(y_offset)].type.equals("stoney block")) {
                blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "tin ore");
              }
            }
          }
        }
        block_x += 20;
      }
      //iron ore
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("stone block") && random(0, 1) <= 0.003) {
            blocks[i][j] = new CustomBlock(i, j, "iron ore");
            float x_offset = 0;
            float y_offset = 0;
            for (int k = 0; k < random(4, 12); k++) {
              //x_offset += round(random(0, 2)-0.5)*2-1;
              //y_offset += round(random(0, 2)-0.5)*2-1;
              x_offset += round(random(-1, 1));
              y_offset += round(random(-1, 1));
              if (i+round(x_offset) < land_width && i+round(x_offset) > 0 && j+round(y_offset) < land_height && j+round(y_offset) > 0 && blocks[i+round(x_offset)][j+round(y_offset)] != null && blocks[i+round(x_offset)][j+round(y_offset)].type.equals("stone block")) {
                blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "iron ore");
              }
            }
          }
        }
        block_x += 20;
      }
      //coal
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("stone block") && random(0, 1) <= 0.003) {
            blocks[i][j] = new CustomBlock(i, j, "coal ore");
            float x_offset = 0;
            float y_offset = 0;
            for (int k = 0; k < random(4, 12); k++) {
              //x_offset += round(random(0, 2)-0.5)*2-1;
              //y_offset += round(random(0, 2)-0.5)*2-1;
              x_offset += round(random(-1, 1));
              y_offset += round(random(-1, 1));
              if (i+round(x_offset) < land_width && i+round(x_offset) > 0 && j+round(y_offset) < land_height && j+round(y_offset) > 0 && blocks[i+round(x_offset)][j+round(y_offset)] != null && blocks[i+round(x_offset)][j+round(y_offset)].type.equals("stone block")) {
                blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "coal ore");
              }
            }
          }
        }
        block_x += 20;
      }
      //pottasium, quartz, thorium ore
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("granite block") && random(0, 1) <= 0.003) {
            int block_name = int(random(0, 3)); //0 = pottasium, 1 = quartz, 2 = thorium
            if (block_name == 0) {
              blocks[i][j] = new CustomBlock(i, j, "potassium ore");
            } else if (block_name == 1) {
              blocks[i][j] = new CustomBlock(i, j, "quartz ore");
            } else {
              blocks[i][j] = new CustomBlock(i, j, "thorium ore");
            }
            float x_offset = 0;
            float y_offset = 0;
            for (int k = 0; k < random(4, 12); k++) {
              //x_offset += round(random(0, 2)-0.5)*2-1;
              //y_offset += round(random(0, 2)-0.5)*2-1;
              x_offset += round(random(-1, 1));
              y_offset += round(random(-1, 1));
              if (i+round(x_offset) < land_width && i+round(x_offset) > 0 && j+round(y_offset) < land_height && j+round(y_offset) > 0 && blocks[i+round(x_offset)][j+round(y_offset)] != null && blocks[i+round(x_offset)][j+round(y_offset)].type.equals("granite block")) {
                if (block_name == 0) {
                  blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "potassium ore");
                } else if (block_name == 1) {
                  blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "quartz ore");
                } else {
                  blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "thorium ore");
                }
              }
            }
          }
        }
        block_x += 20;
      }
      //gems
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("quartz ore") && random(0, 1) <= 0.1) {
            int block_name = int(random(0, 4)); //0 = ruby, 1 = sapphire, 2 = topaz, 3 = amethyst
            if (block_name == 0) {
              blocks[i][j] = new CustomBlock(i, j, "ruby ore");
            } else if (block_name == 1) {
              blocks[i][j] = new CustomBlock(i, j, "sapphire ore");
            } else if (block_name == 2) {
              blocks[i][j] = new CustomBlock(i, j, "topaz ore");
            } else {
              blocks[i][j] = new CustomBlock(i, j, "amethyst ore");
            }
          }
        }
        block_x += 20;
      }
      //boulder
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height; j++) {
          if (blocks[i][j] != null && blocks[i][j].type.equals("dirt") && random(0, 1) <= 0.005) {
            blocks[i][j] = new CustomBlock(i, j, "stone block");
            float x_offset = 0;
            float y_offset = 0;
            for (int k = 0; k < random(8, 20); k++) {
              //x_offset += round(random(0, 2)-0.5)*2-1;
              //y_offset += round(random(0, 2)-0.5)*2-1;
              x_offset = round(random(-2, 1));
              y_offset = round(random(-2, 1));
              if (i+round(x_offset) < land_width && i+round(x_offset) > 0 && j+round(y_offset) < land_height && j+round(y_offset) > 0 && blocks[i+round(x_offset)][j+round(y_offset)] != null && blocks[i+round(x_offset)][j+round(y_offset)].type.equals("dirt")) {
                blocks[i+round(x_offset)][j+round(y_offset)] = new CustomBlock(i+round(x_offset), j+round(y_offset), "stone block");
              }
            }
          }
        }
        block_x += 20;
      }
      //trees
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 12; j < land_height-12; j++) {
          if (blocks[i][j] == null && blocks[i][j+1] != null && blocks[i][j+1].type.equals("dirt") && random(0, 1) <= 0.1) {
            int leaf_counter = 0;
            int tree_height = round(random(4, 12));
            for (int k = 0; k < tree_height; k++) {
              if (i+2 < land_width && i-2 > 0 && j-k > 0) {
                blocks[i][j-k] = new CustomBlock(i, j-k, "wood block");
                if (k > tree_height/2) {
                  blocks[i+1][j-k] = new LeafBlock(block_x+20, (j-k)*20, i+1, j-k, "right");
                  blocks[i-1][j-k] = new LeafBlock(block_x-20, (j-k)*20, i-1, j-k, "left");
                  blocks[i+2][j-k] = new LeafBlock(block_x+40, (j-k)*20, i+2, j-k);
                  blocks[i-2][j-k] = new LeafBlock(block_x-40, (j-k)*20, i-2, j-k);
                }
                leaf_counter++;
              }
            }
            if (i+1 < land_width && i-1 > 0 && j-leaf_counter-1 > 0) {
              blocks[i][j-leaf_counter] = new LeafBlock(block_x, (j-leaf_counter)*20, i, j-leaf_counter, "top");
              blocks[i+1][j-leaf_counter] = new LeafBlock(block_x+20, (j-leaf_counter)*20, i+1, j-leaf_counter);
              blocks[i-1][j-leaf_counter] = new LeafBlock(block_x-20, (j-leaf_counter)*20, i-1, j-leaf_counter);
              blocks[i][j-leaf_counter-1] = new LeafBlock(block_x, (j-leaf_counter)*20-20, i, j-leaf_counter-1);
            }
          }
        }
        block_x += 20;
      }
      //cactus
      block_x = -land_width/2;
      for (int i = 0; i < land_width; i++) {
        for (int j = 0; j < land_height-2; j++) {
          if (blocks[i][j] == null && blocks[i][j+1] != null && blocks[i][j+1].type.equals("sand") && random(0, 1) <= 0.1) {
            String orientation = "up";
            int branch_height = 10;
            int block_list_x = 0;
            int block_list_y = 0;
            //branch
            while (branch_height > 0) {
              for (int k = 0; k < branch_height; k++) {
                if (orientation.equals("up")) {
                  block_list_y--;
                }
                else if (orientation.equals("down")) {
                  block_list_y++;
                }
                else if (orientation.equals("right")) {
                  block_list_x++;
                }
                else if (orientation.equals("left")) {
                  block_list_x--;
                }
                
                String rotation = "h";
                if (orientation.equals("up") || orientation.equals("down")) {
                  rotation = "v";
                }
                if (i+block_list_x >= 0 && i+block_list_x < blocks.length && j+block_list_y >= 0 && j+block_list_y < blocks[0].length) {
                  blocks[i+block_list_x][j+block_list_y] = new CactusBlock(block_x+block_list_x*20, (j+block_list_y)*20, i+block_list_x, j+block_list_y, rotation);
                }
              }
              
              //if (orientation.equals("up")) {
              //  orientation = "right";
              //}
              //else if (orientation.equals("down")) {
              //  orientation = "left";
              //}
              //else if (orientation.equals("right")) {
              //  orientation = "down";
              //}
              //else if (orientation.equals("left")) {
              //  orientation = "up";
              //}
              String[] orientation_list = {"up", "down", "right", "left"};
              orientation = orientation_list[round(random(0, orientation_list.length-1))];
              branch_height -= round(random(0, 3));
            }
            
          }
        }
        block_x += 20;
      }
      
      //lava pits
      block_x = -land_width/2;
      for (int i = 0; i < land_width-10; i++) {
        for (int j = 0; j < land_height-30; j++) {
          if (blocks[i][j] == null && blocks[i][j+1] != null && blocks[i][j+1].type.equals("dried lava block") && random(0, 1) <= 0.05) {
            int pit_depth = round(random(5, 30));
            int pit_width = round(random(4, 10));
            float width_decrease = 0;
            for (int l = -j; l < pit_depth; l++) {
              for (int k = round(width_decrease); k < pit_width-width_decrease; k++) {
                blocks[i+k][j+l] = null;
              }
              //width_decrease = 1;
              if (l > 0) {
                width_decrease += 0.1;
              }
              //width_decrease++;
            }
            for (int k = 0; k < pit_width; k++) {
              blocks[i+k][j] = new LavaBlock(block_x+k*20, j*20, i+k, j);
            }
          }
        }
        block_x += 20;
      }
      
      ArrayList<Light> lights = new ArrayList<Light>();
      for (int i = max(0, round(player.x/20)-40); i < min(round(player.x/20)+40, blocks.length); i++) {
        for (int j = max(0, round(player.y/20)-40); j < min(round(player.y/20)+40, blocks[i].length); j++) {
          if (blocks[i][j] != null && blocks[i][j].light != null) {
            lights.add(blocks[i][j].light);
          }
        }
      }
      for (int i = max(0, round(player.x/20)-40); i < min(round(player.x/20)+40, blocks.length); i++) {
        int min_y_list_pos = 100000;
        for (int j = 0; j < blocks[i].length; j++) {
          if (blocks[i][j] != null) {
            min_y_list_pos = min(min_y_list_pos, j);
          }
        }
        if (i > 1 && i < blocks.length-1) {
          for (int j = 0; j < min_y_list_pos; j++) {
            if (blocks[i-1][j] != null) {
              lights.add(new Light(blocks[i-1][j].x, blocks[i-1][j].y, 100, color(255)));
            }
            if (blocks[i+1][j] != null) {
              lights.add(new Light(blocks[i+1][j].x, blocks[i+1][j].y, 100, color(255)));
            }
          }
        }
        lights.add(new Light(blocks[i][min_y_list_pos].x, blocks[i][min_y_list_pos].y, 100, color(255)));
      }
      for (Block[] block_row : blocks) {
        for (Block block : block_row) {
          if (block != null) {
            block.update(lights);
          }
        }
      }
      state = "play";
      fill(0);
      textFont(createFont("Arial", 15, true), 15);
      textAlign(RIGHT);
    }
  }

  void update() {
    if (state.equals("menu")) {
      image(main_menu_screen, width/2, height/2);
      Button button_continue = new Button(width/2, height/2, "button_continue.png", "button_continue_hovered.png");
      Button button_new_game = new Button(width/2, height*0.65, "button_new_game.png", "button_new_game_hovered.png");
      Button button_exit = new Button(width/2, height*0.9, "button_exit.png", "button_exit_hovered.png");
      button_continue.display();
      button_new_game.display();
      button_exit.display();
      if (button_continue.clicked) {
        state = "load_world_file";
      } else if (button_new_game.clicked) {
        state = "load_1";
      } else if (button_exit.clicked) {
        exit();
      }
    } else if (state.equals("play")) {
      textAlign(RIGHT);
      background(25, 120, 250);
      pushMatrix();
      
      //float player_biome_offset = (player.x+400)%(biome_width*20);
      if (player.x+400 < biome_width*19) {
        //image(biome_backgrounds[0], width/2+map((player.x+400)%(biome_width*20), 0, biome_width*20, -480, 480), height/2);
        //image(biome_backgrounds[0], width/2+(((player.x+400)/(biome_width*20))*960)-480, height/2);
        //image(biome_backgrounds[0], width/2+(((400)/(biome_width*20))*960)-480, height/2);
        //image(biome_backgrounds[0], ((player.x+400)*(1/(biome_width*20)))*960, height/2);
        //image(biome_backgrounds[0], width/2+(((player.x+400)/(1*(biome_width*20)))*960)-480, height/2);
        image(biome_backgrounds[0], constrain(width/2-int((((player.x)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      } else if (player.x+400 < biome_width*21) {
        image(biome_backgrounds[0], constrain(width/2-int((((player.x)+400)/(biome_width*20))*960)+480, 0, width), height/2);
        tint(255, map(player.x+400-biome_width*19, 0, biome_width*2, 1, 254));
        image(biome_backgrounds[1], constrain(width/2-int((((player.x-biome_width*20)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      } else if (player.x+400 < biome_width*39) {
        image(biome_backgrounds[1], constrain(width/2-int((((player.x-biome_width*20)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      } else if (player.x+400 < biome_width*41) {
        image(biome_backgrounds[1], constrain(width/2-int((((player.x-biome_width*20)+400)/(biome_width*20))*960)+480, 0, width), height/2);
        tint(255, map(player.x+400-biome_width*39, 0, biome_width*2, 1, 254));
        image(biome_backgrounds[2], constrain(width/2-int((((player.x-biome_width*40)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      } else if (player.x+400 < biome_width*59) {
        image(biome_backgrounds[2], constrain(width/2-int((((player.x-biome_width*40)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      } else if (player.x+400 < biome_width*61) {
        image(biome_backgrounds[2], constrain(width/2-int((((player.x-biome_width*40)+400)/(biome_width*20))*960)+480, 0, width), height/2);
        tint(255, map(player.x+400-biome_width*59, 0, biome_width*2, 1, 254));
        image(biome_backgrounds[3], constrain(width/2-int((((player.x-biome_width*60)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      } else {
        image(biome_backgrounds[3], constrain(width/2-int((((player.x-biome_width*60)+400)/(biome_width*20))*960)+480, 0, width), height/2);
      }
      noTint();
      
      translate(width/2-player.x, height/2-player.y);
      int render_distance = 25;
      //finds all the light sources
      ArrayList<Light> lights = new ArrayList<Light>();
      for (Particle particle : particles) {
        if (particle.light_level > 0 && particle.fuse > 0) {
          lights.add(new Light(particle.x, particle.y, particle.light_level, particle.theColor));
        }
      }
      for (int i = max(0, round((player.x+200)/20)-render_distance); i < min(round((player.x+200)/20)+render_distance, blocks.length); i++) {
        for (int j = max(0, round(player.y/20)-render_distance); j < min(round(player.y/20)+render_distance, blocks[i].length); j++) {
          if (blocks[i][j] != null && blocks[i][j].light != null) {
            lights.add(blocks[i][j].light);
          }
        }
      }
      
      //the 'sun'
      //render_distance *= 1.5;
      for (int i = max(0, round(player.x/20)-render_distance+player_offset); i < min(round(player.x/20)+render_distance+player_offset, blocks.length); i++) {
        int min_y_list_pos_blocks = 1000000;
        //for (int j = 0; j < blocks[i].length-1; j++) {
        for (int j = max(0, round(player.y/20)-render_distance); j < min(round(player.y/20)+render_distance, blocks[i].length); j++) {
          if (blocks[i][j] != null) {
            min_y_list_pos_blocks = min(min_y_list_pos_blocks, j);
          }
        }
        if (min_y_list_pos_blocks == 1000000) {
          continue;
        }
        
        for (int j = max(0, round(player.y/20)-render_distance); j < min_y_list_pos_blocks; j++) {
          try {
            if (i >= 1 && (blocks[i-1][j] != null)) {
              lights.add(new Light(blocks[i-1][j].x, blocks[i-1][j].y, 100, color(255)));
            }
          } catch (ArrayIndexOutOfBoundsException e) {
            println("i: " + i + " j: " + j);
          }
          if (i <= blocks.length-2 && (blocks[i+1][j] != null)) {
            lights.add(new Light(blocks[i+1][j].x, blocks[i+1][j].y, 100, color(255)));
          }
        }
        
        lights.add(new Light(blocks[i][min_y_list_pos_blocks].x, blocks[i][min_y_list_pos_blocks].y, 100, color(255)));
      }
      //render_distance /= 1.5;
     
      //updates all the blocks evenly over 60 ticks
      for (int i = max(0, round((player.x/20)+player_offset)+round(2*render_distance*(block_update_counter/max_block_update_counter)) - render_distance); i < min(blocks.length, round((player.x/20)+player_offset) + round(2*render_distance*((block_update_counter+1)/max_block_update_counter)) - render_distance); i++) {
        for (int j = max(0, round(player.y/20)-render_distance); j < min(round(player.y/20)+render_distance, blocks[i].length); j++) {
          if (background_walls[i][j] != null) {
            background_walls[i][j].update(lights);
          }
          if (blocks[i][j] != null) {
            blocks[i][j].update(lights);
          }
        }
      }
      for (int i = max(0, round(player.x/20)-render_distance+player_offset); i < min(round(player.x/20)+render_distance+player_offset, blocks.length); i++) {
        for (int j = max(0, round(player.y/20)-render_distance); j < min(round(player.y/20)+render_distance, blocks[i].length); j++) {
          if (background_walls[i][j] != null) {
            background_walls[i][j].display();
          }
        }
      }
      for (int i = max(0, round(player.x/20)-render_distance+player_offset); i < min(round(player.x/20)+render_distance+player_offset, blocks.length); i++) {
        for (int j = max(0, round(player.y/20)-render_distance); j < min(round(player.y/20)+render_distance, blocks[i].length); j++) {
          if (blocks[i][j] != null) {
            blocks[i][j].display(lights);
          }
        }
      }
      
      if (block_update_counter == 0) {
        float average_fps = 0;
        for (float recent_fps_value : recent_fps) {
          average_fps += recent_fps_value/recent_fps.size();
        }
        //if (average_fps < 55) {
        //  max_block_update_counter++;
        //} else {
        //  max_block_update_counter--;
        //}
        if (abs(average_fps-55) > 5) {
          max_block_update_counter = constrain(max_block_update_counter+round((55-average_fps)/5), 10, 10);
        }
        block_update_counter = max_block_update_counter;
      }
      block_update_counter--;
      for (int i = enemies.size() - 1; i > 0; i--) {
        if (!enemies.get(i).is_dead && dist(player.x, player.y, enemies.get(i).location.x, enemies.get(i).location.y) < 20*40) {
          enemies.get(i).update(player);
          enemies.get(i).display();
        } else if (enemies.get(i).is_dead) {
          enemies.remove(i);
        }
      }
      for (int i = enemy_bullets.size() - 1; i > 0; i--) {
        if (enemy_bullets.get(i).should_be_destroyed()) {
          enemy_bullets.remove(i);
        } else {
          enemy_bullets.get(i).update();
        }
      }
      player.update(lights);
      //for (int i = enemies.size() - 1; i > 0; i--) {
      //  player.checkAttack(enemies.get(i));
      //}
      for (Particle particle : particles) {
        particle.update();
      }
      text("FPS: " + str(round(frameRate)), width, 12);
      text("TPU: " + max_block_update_counter, width, 25);
      text("Health: " + player.health, width, 37);
      recent_fps.append(frameRate);
      while (recent_fps.size() >= 50) {
        recent_fps.remove(0);
      }
      
      forest_enemies = 0;
      cave_enemies = 0;
      desert_enemies = 0;
      wasteland_enemies = 0;
      for (Enemy enemy : enemies) {
        if (!enemy.is_dead) {
          if (enemy.type.equals("goblin") || enemy.type.equals("orc") || enemy.type.equals("forest wizard") || enemy.type.equals("phantom") || enemy.type.equals("cyclops") || enemy.type.equals("cyclops tower")) {
            forest_enemies++;
          } else if (enemy.type.equals("skeleton") || enemy.type.equals("skeleton knight") || enemy.type.equals("necromancer")) {
            cave_enemies++;
          } else if (enemy.type.equals("angry cactus") || enemy.type.equals("sand wraith")) {
            desert_enemies++;
          } else if (enemy.type.equals("shear back beetle") || enemy.type.equals("living bomb") || enemy.type.equals("living boulder")) {
            wasteland_enemies++;
          } else {
            println("Unidentified enemy: " + enemy.type);
            exit();
          }
        }
      }
      
      if (player.x + player_offset < biome_width * 20) {                        //forest
        if (pow( spawn_speed, forest_enemies) > random(spawn_speed)) {
          float spawn = random(1);
          float enemy_x = random(-player_offset, biome_width*20-player_offset);
          int enemy_y = 1000000;
          for (int j = 0; j < blocks[0].length-1; j++) {
            if (blocks[round((enemy_x+player_offset)/20)][j] != null) {
              enemy_y = min(enemy_y, j*20);
            }
          }
          
          if (spawn > 0.9) {
            enemies.add(new ForestWizard(enemy_x, enemy_y, blocks));
          } else if (spawn > 0.3) {
            enemies.add(new Goblin(enemy_x, enemy_y, blocks));
          } else if (spawn > 0.1) {
            enemies.add(new Orc(enemy_x, enemy_y, blocks));
          } else {
            enemies.add(new Phantom(enemy_x, random(-player_offset, biome_width*20-player_offset)));
          }
          //enemies.add(new Cyclops(enemy_x, enemy_y, blocks));
        }
      } else if (player.x + player_offset < 2 * biome_width * 20) {             //undergorund
        if (pow( spawn_speed, cave_enemies) > random(spawn_speed)) {
          float spawn = random(1);
          float enemy_x = random(biome_width*20-player_offset, 2*biome_width*20-player_offset);
          int enemy_y = 1000000;
          for (int j = 0; j < blocks[0].length-1; j++) {
            if (blocks[round((enemy_x+player_offset)/20)][j] != null) {
              enemy_y = min(enemy_y, j*20);
            }
          }
          
          if (spawn > 0.4) {
            enemies.add(new Skeleton(enemy_x, enemy_y, blocks));
          } else if (spawn > 0.1) {
            enemies.add(new SkeletonKnight(enemy_x, enemy_y, blocks));
          } else {
            enemies.add(new Necromancer(enemy_x, enemy_y, blocks));
          }
        }
      } else if (player.x + player_offset < 3 * biome_width * 20) {            //desert
        if (pow( spawn_speed, desert_enemies) > random(spawn_speed)) {
          float spawn = random(1);
          float enemy_x = random(2*biome_width*20-player_offset, 3*biome_width*20-player_offset);
          int enemy_y = 1000000;
          for (int j = 0; j < blocks[0].length-1; j++) {
            if (blocks[round((enemy_x+player_offset)/20)][j] != null) {
              enemy_y = min(enemy_y, j*20);
            }
          }
          
          if (spawn > 0.5) {
            enemies.add(new AngryCactus(enemy_x, enemy_y, blocks));
          } else {
            enemies.add(new SandWraith(enemy_x, enemy_y));
          }
        }
      } else if (player.x + player_offset < 4 * biome_width * 20) {              //wasteland
        if (pow( spawn_speed, wasteland_enemies) > random(spawn_speed)) {
          float spawn = random(1);
          float enemy_x = random(3*biome_width*20-player_offset, 4*biome_width*20-player_offset);
          int enemy_y = 1000000;
          for (int j = 0; j < blocks[0].length-1; j++) {
            if (blocks[round((enemy_x+player_offset)/20)][j] != null) {
              enemy_y = min(enemy_y, j*20);
            }
          }
          
          if (spawn > 0.5) {
            enemies.add(new ShearBackBeetle(enemy_x, enemy_y, blocks));
          } else if (spawn > 0.25) {
            enemies.add(new LivingBomb(enemy_x, enemy_y, blocks));
          } else {
            enemies.add(new LivingBoulder(enemy_x, enemy_y, blocks));
            println(enemy_x, enemy_y);
          }
        }
      }
    } else if (state.equals("load_1")) {
      loadGame(1);
    } else if (state.equals("load_2")) {
      loadGame(2);
    } else if (state.equals("load_3")) {
      loadGame(3);
    } else if (state.equals("load_world_file")) {
      load_world("world", 0, 0);
    }
  }
  
  void save_world() {
    JSONArray world = new JSONArray();
    int block_counter = 0;
    for (int i = 0; i < blocks.length; i++) {
      for (int j = 0; j < blocks[i].length; j++) {
        if (blocks[i][j] != null) {
          JSONObject current_block = new JSONObject();
          //current_block.setFloat("x", blocks[i][j].x);
          //current_block.setFloat("y", blocks[i][j].y);
          current_block.setString("data_type", "block");
          current_block.setInt("x_list_pos", blocks[i][j].x_list_pos);
          current_block.setInt("y_list_pos", blocks[i][j].y_list_pos);
          current_block.setString("type", blocks[i][j].type);
          world.setJSONObject(block_counter, current_block);
          block_counter++;
        }
      }
    }
    for (int i = 0; i < background_walls.length; i++) {
      for (int j = 0; j < background_walls[i].length; j++) {
        if (background_walls[i][j] != null) {
          JSONObject current_block = new JSONObject();
          current_block.setString("data_type", "background wall");
          current_block.setFloat("x", background_walls[i][j].x);
          current_block.setFloat("y", background_walls[i][j].y);
          current_block.setInt("x_list_pos", background_walls[i][j].x_list_pos);
          current_block.setInt("y_list_pos", background_walls[i][j].y_list_pos);
          current_block.setString("img_name", background_walls[i][j].img_name);
          current_block.setString("block_name", background_walls[i][j].type);
          world.setJSONObject(block_counter, current_block);
          block_counter++;
        }
      }
    }
    saveJSONArray(world, "data/world.json");
  }
  
  void load_world(String world_name, int x_off, int y_off) { //blocks
    state = "play";
    JSONArray world;
    world = loadJSONArray("data/"+world_name+".json");
    for (int i = 0; i < world.size(); i++) {
      JSONObject current_block = world.getJSONObject(i);
      if (current_block.getString("data_type").equals("block")) {
        create_block(current_block.getInt("x_list_pos")+x_off, current_block.getInt("y_list_pos")+y_off, current_block.getString("type"));
      } else {
        background_walls[current_block.getInt("x_list_pos")+x_off][current_block.getInt("y_list_pos")+y_off] = new CustomBackgroundWall(current_block.getFloat("x")+x_off*20, current_block.getFloat("y")+y_off*20, current_block.getInt("x_list_pos")+x_off, current_block.getInt("y_list_pos")+y_off, current_block.getString("img_name"), current_block.getString("block_name"));
      }
    }
    for (int i = 0; i < blocks.length; i++) {
      for (int j = 0; j < blocks[i].length; j++) {
        if (blocks[i][j] != null) {
          ArrayList lights = new ArrayList<Light>();
          blocks[i][j].update(lights);
        }
      }
    }
  }
  
  void create_block(int _x_list_pos, int _y_list_pos, String block_name) {
    if (block_name.equals("altar")) {
      blocks[_x_list_pos][_y_list_pos] = new Altar(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos);
    } else if (block_name.equals("cactus")) {
      blocks[_x_list_pos][_y_list_pos] = new CactusBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos, "v");
    } else if (block_name.equals("dirt")) {
      blocks[_x_list_pos][_y_list_pos] = new DirtBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos);
    } else if (block_name.equals("dried lava block")) {
      blocks[_x_list_pos][_y_list_pos] = new DriedLavaBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos);
    } else if (block_name.equals("lava")) {
      blocks[_x_list_pos][_y_list_pos] = new LavaBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos);
    } else if (block_name.equals("primary leaf block")) {
      blocks[_x_list_pos][_y_list_pos] = new LeafBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos, "top");
    } else if (block_name.equals("secondary leaf block")) {
      blocks[_x_list_pos][_y_list_pos] = new LeafBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos);
    } else if (block_name.equals("sand")) {
      blocks[_x_list_pos][_y_list_pos] = new SandBlock(_x_list_pos*20-400, _y_list_pos*20, _x_list_pos, _y_list_pos);
    } else {
      blocks[_x_list_pos][_y_list_pos] = new CustomBlock(_x_list_pos, _y_list_pos, block_name);
    }
  }
}